/*
 * Catweasel -- Advanced Floppy Controller
 * Linux device driver
 * Low-level routines
 *
 * Copyright (C) 1998 Michael Krause
 * Modifications by Timothy Mann for use with cw2dmk
 * $Id: catweasl.c,v 1.8 2001/07/15 01:25:59 mann Exp $
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#define CHECK_DISK_CHANGED 0  /* older drives don't provide this signal */

#ifdef __KERNEL__
#include <asm/io.h>
#endif

#include "cwfloppy.h"
#include <sys/time.h>

#define CREG(c) (c->private[0])
#define MEMSIZE 131072

/* Await particular bit value(s) in the status register.  Return 1 if
   it appears, 0 if 5 seconds pass and it doesn't.  Bits that are 1
   in "clear" must be 0; bits that are 1 in "set" must be 1; others
   are don't-cares. */
int catweasel_await_status(catweasel_contr *c, int clear, int set)
{
  int i;
  struct timeval tv1, tv2;
  int iobase = c->iobase;

  gettimeofday(&tv1, NULL);
  for (;;) {
    i = 1000000;
    while (i--) {
      int v = inb(iobase+2);
      if ((v & clear) == 0 && (v & set) == set) return 1;
    }
    gettimeofday(&tv2, NULL);
    if ((tv2.tv_sec - tv1.tv_sec)*1000000 + tv2.tv_usec - tv2.tv_usec
	> 5000000) return 0;
  }
}

/* Trivial memory test.  Returns number of seemingly good bytes. */
int catweasel_memtest(catweasel_contr *c)
{
  int i;
  int iobase = c->iobase;

  inb(iobase+1);  /* reset RAM pointer */
  for (i=0; i<MEMSIZE; i++) {
    outb(i%4093, iobase);
  }
  inb(iobase+1);  /* reset RAM pointer */
  for (i=0; i<MEMSIZE; i++) {
    if (inb(iobase) != (unsigned char)(i%4093)) {
      return i;
    }
  }
  inb(iobase+1);  /* reset RAM pointer */
  for (i=0; i<MEMSIZE; i++) {
    outb(~(i%4093), iobase);
  }
  inb(iobase+1);  /* reset RAM pointer */
  for (i=0; i<MEMSIZE; i++) {
    if (inb(iobase) != (unsigned char)~(i%4093)) {
      return i;
    }
  }
  return i;
}

#if DEBUG5
/* Fill Catweasel's memory with specified value, and check it is there. */
static void catweasel_fillmem(catweasel_drive *d, unsigned char byte)
{
  int i;
  int iobase = d->contr->iobase;

  inb(iobase+1);  /* reset RAM pointer */
  for (i=0; i<MEMSIZE; i++) {
    outb(byte, iobase);
  }
  inb(iobase+1);  /* reset RAM pointer */
  for (i=0; i<MEMSIZE; i++) {
    if (inb(iobase) != byte) {
      printf("error at %d\n", i);
    }
  }
}
#endif /*DEBUG5*/

static __inline__ void CWSetCReg(catweasel_contr *c, unsigned char clear, unsigned char set)
{
    CREG(c) = (CREG(c) & ~clear) | set;
    outb(CREG(c), c->iobase + 2);
}

static void CWTriggerStep(catweasel_contr *c)
{
    CWSetCReg(c, 1<<0, 0);
    c->msdelay(3);
    CWSetCReg(c, 0, 1<<0);
    c->msdelay(3);
}

void catweasel_init_controller(catweasel_contr *c)
{
    int i, j;

    if(!c->iobase)
	return;

    CREG(c) = 255;

    /* select all drives, step inside */
    CWSetCReg(c, 1<<1 | 1<<4 | 1<<5, 0);
    for(i=0;i<3;i++)
	CWTriggerStep(c);

    for(i=0;i<2;i++) {
	c->drives[i].number = i;
	c->drives[i].contr = c;
	c->drives[i].diskindrive = 0;
	
	/* select only the respective drive, step to track 0 */
	CWSetCReg(c, 0, 1<<4 | 1<<5 | 1<<1);
	CWSetCReg(c, 1<<(5-i), 0);
	
	for(j=0;j<85 && inb(c->iobase+2) & 16;j++)
	    CWTriggerStep(c);
	
	if(j < 85) {
	    c->drives[i].type = 1;
	    c->drives[i].track = 0;
	} else {
	    c->drives[i].type = 0;
	}
    }
    
    CWSetCReg(c, 0, 1<<4 | 1<<5); /* deselect all drives */
}

void catweasel_free_controller(catweasel_contr *c)
{
    if(!c->iobase)
	return;

    /* all motors off, deselect all drives */
    CWSetCReg(c, 0, 1<<3 | 1<<7 | 1<<4 | 1<<5);
}

void catweasel_select(catweasel_contr *c, int dr0, int dr1)
{
    CWSetCReg(c, dr0 << 5 | dr1 << 4, (!dr0) << 5 | (!dr1) << 4);
}

static const unsigned char motorbits[2] = { 3, 7 };

void catweasel_set_motor(catweasel_drive *d, int on)
{
    int mask = 1 << motorbits[d->number];
    
    if(on)
	CWSetCReg(d->contr, mask, 0);
    else
	CWSetCReg(d->contr, 0, mask);
}

void catweasel_seek(catweasel_drive *d, int t)
{
    int x;

    if(t >= 80) {
	return;
    }

    x = t - d->track;
    if(!x)
	return;

    if(x >= 0) {
	CWSetCReg(d->contr, 1<<1, 0);
    } else {
	CWSetCReg(d->contr, 0, 1<<1);
	x = -x;
    }

    while(x--)
	CWTriggerStep(d->contr);

    d->track = t;
}

#if CHECK_DISK_CHANGED
int catweasel_disk_changed(catweasel_drive *d)
{
    int ot, t, changed = 0;

    if(inb(d->contr->iobase+2) & 32) {
	if(!d->diskindrive) {
	    /* first usage of this drive, issue disk change */
	    d->diskindrive = 1;
	    changed = 1;
	} else {
	    /* there's still a disk in there, no change detected. */
	}
    } else {
	ot = d->track;
	if(ot == 79)
	    t = 78;
	else
	    t = ot + 1;
	
	catweasel_seek(d, t);
	catweasel_seek(d, ot);
	
	if(!(inb(d->contr->iobase+2) & 32)) {
	    if(!d->diskindrive) {
		/* drive still empty, nothing has happened */
	    } else {
		/* disk has been removed */
		d->diskindrive = 0;
		changed = 1;
	    }
	} else {
	    /* disk has been inserted */
	    d->diskindrive = 1;
	    changed = 1;
	}
    }

    return changed;
}
#endif

int catweasel_write_protected(catweasel_drive *d)
{
    return !(inb(d->contr->iobase+2) & 8);
}

int catweasel_read(catweasel_drive *d, int side, int clock, int time, int idx)
{
    int iobase = d->contr->iobase;

#if CHECK_DISK_CHANGED
    if (!catweasel_await_status(d->contr, 0, 1<<5)) return 0;
#endif

    CWSetCReg(d->contr, 1<<2, (!side)<<2); /* set disk side */

    inb(iobase+1);              /* reset RA */
    outb(clock*128, iobase+3);  /* select clock */

    inb(iobase+1);              /* reset RA */
    inb(iobase+0);              /* RA++ */
    inb(iobase+0);              /* RA++ */
    if (idx) {
      outb(128, iobase+3);      /* store index pulse in high-order bit */
    } else {
      outb(0, iobase+3);        /* don't store index pulse */
    }
    inb(iobase+1);              /* reset RA */

    if (time < 0) {
      /* read index hole to index hole */
      inb(iobase+5);                 
      /* wait for read to start */
      if (!catweasel_await_status(d->contr, 1<<0, 0)) return 0;
      /* wait for read to end */
      if (!catweasel_await_status(d->contr, 0, 1<<0)) return 0;
    } else {
      inb(iobase+7);                 /* start reading immediately */
      d->contr->msdelay(time);       /* wait the prescribed time */
      outb(0, iobase+1);             /* stop reading, don't reset RA */
    }
    outb(128, iobase+0);        /* add data end mark */
    outb(128, iobase+0);
    outb(0, iobase+0);
    outb(128, iobase+0);
    outb(0, iobase+0);
    outb(128, iobase+0);
    outb(0, iobase+0);

    inb(iobase+1);              /* reset RA */
    inb(iobase+0);              /* ignore first byte */

    return 1;
}

int catweasel_write(catweasel_drive *d, int side, int clock, int time)
{
    int iobase = d->contr->iobase;

#if CHECK_DISK_CHANGED
    if (!catweasel_await_status(d->contr, 0, 1<<5)) return 0;
#endif

    CWSetCReg(d->contr, 1<<2, (!side)<<2); /* set disk side */

    inb(iobase+1);              /* reset RA */
    outb(clock*128, iobase+3);  /* select clock */
    inb(iobase);                /* RA++ */
    outb(128, iobase+3);        /* enable writing */
    inb(iobase);                /* RA++ */
    inb(iobase);                /* RA++ */
    inb(iobase);                /* RA++ */
    inb(iobase);                /* RA++ */
    inb(iobase);                /* RA++ */
    inb(iobase);                /* RA++ (=7) */
    if (time < 0) {
      /* write from index hole to index hole */
      outb(0, iobase+7);             
      /* wait for write to start */
      if (!catweasel_await_status(d->contr, 1<<1, 0)) return 0;
      /* wait for write to end */
      if (!catweasel_await_status(d->contr, 0, 1<<1)) return 0;
    } else {
      outb(0, iobase+5);             /* start writing immediately */
      d->contr->msdelay(time);       /* wait for prescribed time */
    }
    inb(iobase+1);              /* stop writing and reset RA */
    return 1;
}

#ifndef __KERNEL__
/* Should be modified to time out after a while */
int catweasel_await_index(catweasel_drive *d)
{
#if CHECK_DISK_CHANGED
    if (!catweasel_await_status(d->contr, 0, 1<<5)) return 0;
#endif
    if (!catweasel_await_status(d->contr, 0, 1<<6)) return 0;
    if (!catweasel_await_status(d->contr, 1<<6, 0)) return 0;
    return 1;
}
#endif

